import * as React from 'react';
import { getFormValues } from '@theoryofnekomata/formxtra';
import fetchPonyfill from 'fetch-ponyfill';
import { NextPage as DefaultNextPage } from 'next';
import {
	NextApiResponse,
	NextApiRequest,
	ENCTYPE_APPLICATION_JSON,
	ENCTYPE_MULTIPART_FORM_DATA,
	ENCTYPE_X_WWW_FORM_URLENCODED,
} from './common';

const FormDerivedElementComponent = 'form' as const;

type FormDerivedElement = HTMLElementTagNameMap[typeof FormDerivedElementComponent];

type BaseProps = React.HTMLProps<FormDerivedElement>;

type EncTypeSerializer = (data: unknown) => string;

type EncTypeSerializerMap = Record<string, EncTypeSerializer>;

const DEFAULT_ENCTYPE_SERIALIZERS: EncTypeSerializerMap = {
	[ENCTYPE_APPLICATION_JSON]: (data: unknown) => JSON.stringify(data),
};

type GetFormValuesOptions = Parameters<typeof getFormValues>[1];

export interface FormProps extends Omit<BaseProps, 'action'> {
	action?: string;
	clientAction?: string;
	clientHeaders?: HeadersInit;
	clientMethod?: BaseProps['method'];
	invalidate?: (...args: unknown[]) => unknown;
	refresh?: (response: Response) => void;
	encTypeSerializers?: EncTypeSerializerMap;
	responseEncType?: string;
	serializerOptions?: GetFormValuesOptions;
}

export const useResponse = (res: NextApiResponse) => {
	const [response, setResponse] = React.useState<Response | undefined>(
		res.body ? new Response(res.body as unknown as BodyInit) : undefined,
	);

	const invalidate = React.useCallback(() => {
		setResponse(undefined);
	}, []);

	const refresh = React.useCallback((newResponse: Response) => {
		setResponse(newResponse);
	}, []);

	return React.useMemo(() => ({
		response,
		refresh,
		invalidate,
	}), [
		response,
		refresh,
		invalidate,
	]);
};

interface SerializeBodyParams {
	form: HTMLFormElement,
	encType: string,
	serializers: EncTypeSerializerMap,
	options?: GetFormValuesOptions,
}

const serializeBody = (params: SerializeBodyParams) => {
	const {
		form,
		encType,
		serializers,
		options,
	} = params;

	if (encType === ENCTYPE_MULTIPART_FORM_DATA) {
		// type error when provided a submitter element for some reason...
		const FormDataUnknown = FormData as unknown as {
			new(formElement?: HTMLElement, submitter?: HTMLElement): BodyInit;
		};
		return new FormDataUnknown(form, options?.submitter);
	}

	if (encType === ENCTYPE_X_WWW_FORM_URLENCODED) {
		return new URLSearchParams(form);
	}

	if (typeof serializers[encType] === 'function') {
		return serializers[encType](getFormValues(form, options));
	}

	throw new Error(`Unsupported encType: ${encType}`);
};

export const Form = React.forwardRef<FormDerivedElement, FormProps>(({
	children,
	onSubmit,
	action,
	method = 'get',
	clientAction = action,
	clientMethod = method,
	clientHeaders,
	encType = ENCTYPE_MULTIPART_FORM_DATA,
	invalidate,
	refresh,
	encTypeSerializers = DEFAULT_ENCTYPE_SERIALIZERS,
	responseEncType = ENCTYPE_APPLICATION_JSON,
	serializerOptions,
	...etcProps
}, forwardedRef) => {
	const handleSubmit: React.FormEventHandler<FormDerivedElement> = async (event) => {
		event.preventDefault();
		const nativeEvent = event.nativeEvent as unknown as { submitter?: HTMLElement };

		if (clientAction) {
			invalidate?.();
			const { fetch } = fetchPonyfill();
			const headers: HeadersInit = {
				...(clientHeaders ?? {}),
				Accept: responseEncType,
			};
			if (encType !== ENCTYPE_MULTIPART_FORM_DATA) {
				// browser automatically generates content-type header for multipart/form-data
				(headers as unknown as Record<string, string>)['Content-Type'] = encType;
			}

			const fetchInit: RequestInit = {
				method: clientMethod.toUpperCase(),
				headers,
			};

			if (!['GET', 'HEAD'].includes(clientMethod.toUpperCase())) {
				fetchInit.body = serializeBody({
					form: event.currentTarget,
					encType,
					serializers: encTypeSerializers,
					options: {
						...serializerOptions,
						submitter: nativeEvent.submitter,
					},
				});
			}
			const response = await fetch(clientAction, fetchInit);
			refresh?.(response);
		}

		onSubmit?.(event);
	};

	// TODO how to display put/patch method in form? HTML only supports get/post
	//   > throw error when not get/post
	// TODO handle "dialog" method as invalid

	const serverMethodRaw = method.toLowerCase();
	const serverMethod = serverMethodRaw === 'get' ? 'get' : 'post';

	return (
		<FormDerivedElementComponent
			{...etcProps}
			ref={forwardedRef}
			onSubmit={handleSubmit}
			action={action}
			method={serverMethod}
			encType={ENCTYPE_MULTIPART_FORM_DATA}
		>
			{children}
		</FormDerivedElementComponent>
	);
});

Form.displayName = 'Form' as const;

Form.defaultProps = {
	action: undefined,
	clientAction: undefined,
	clientHeaders: undefined,
	clientMethod: undefined,
	encTypeSerializers: DEFAULT_ENCTYPE_SERIALIZERS,
	invalidate: undefined,
	refresh: undefined,
	responseEncType: ENCTYPE_APPLICATION_JSON,
	serializerOptions: undefined,
};

export type NextPage<T = NonNullable<unknown>, U = T> = DefaultNextPage<
	T & {
		res: NextApiResponse;
		req: NextApiRequest;
	}, U
>
